/**
 *                _ _   _               _
 * __      ___ __(_) |_| |_ ___ _ __   | |__  _   _
 * \ \ /\ / / '__| | __| __/ _ \ '_ \  | '_ \| | | |
 *  \ V  V /| |  | | |_| ||  __/ | | | | |_) | |_| |
 *   \_/\_/ |_|  |_|\__|\__\___|_| |_| |_.__/ \__, |
 *                                            |___/
 *  ____                  __     __               _
 * / ___|_   _____ _ __   \ \   / /__   __ _  ___| |
 * \___ \ \ / / _ \ '_ \   \ \ / / _ \ / _` |/ _ \ |
 *  ___) \ V /  __/ | | |   \ V / (_) | (_| |  __/ |
 * |____/ \_/ \___|_| |_|    \_/ \___/ \__, |\___|_|
 *                                     |___/
 * Licensed under the GPLv2 License, Version 2.0 (the "License");
 * Copyright (c) Sven Vogel
 */

use crate::FloatOrChar::{Char, Float};
use std::collections::{HashMap, HashSet};
use std::fmt::Debug;
use std::usize;

/// pretty useless enum that either contains a char or
/// a f32
#[derive(Debug)]
enum FloatOrChar {
    Float(f64),
    Char(char),
}

impl PartialEq for FloatOrChar {
    /// only returns true if
    /// self and other of the same enum variant and their wrapped values match
    fn eq(&self, other: &Self) -> bool {
        return match self {
            Float(x) => {
                if let Float(y) = *other {
                    // compare floats with tolerance
                    (*x - y).abs() < f64::EPSILON
                } else {
                    false // other is not a Float32
                }
            }
            Char(x) => {
                if let Char(y) = *other {
                    *x == y
                } else {
                    false // other is not a Char
                }
            }
        };
    }

    /// opposite of equals
    fn ne(&self, other: &Self) -> bool {
        !self.eq(other)
    }
}

/// count all duplicate values in an array
/// equivalence is determined by the [`PartialEq`] trait
fn duplicates<T>(arr: &[T]) -> usize
where
    T: PartialEq + Debug,
{
    let mut total_dups = 0;

    for x in 0..arr.len() {
        let to_cmp = &arr[x];
        let mut count = 0;

        for y in (x + 1)..arr.len() {
            if &arr[y] == to_cmp {
                count += 1;
            }
        }

        total_dups += count;

        if count > 0 {
            println!("Element {:?} was duplicated {} time(s)", to_cmp, count);
        }
    }

    total_dups
}

/// read some data from stdin and stores it into a vector.
/// the reading of data stops if the input is neither a float nor a single character.
/// input line are trimmed
fn read_array() -> Vec<FloatOrChar> {
    // temporary string buffer for reading
    let mut buf = String::new();
    // storage for our parsed array
    let mut elems = Vec::<FloatOrChar>::new();

    println!("Enter some floats or characters, one per line:");

    loop {
        // read some characters
        std::io::stdin().read_line(&mut buf).unwrap();

        // parse the data
        if let Ok(float) = buf.trim().parse::<f64>() {
            elems.push(Float(float));
        } else if let Ok(char) = buf.trim().parse::<char>() {
            elems.push(Char(char));
        } else {
            break;
        }

        // empty string
        buf.clear();
    }

    elems
}

fn main() {
    let arr = &read_array()[..];

    println!("Duplicates: {}", duplicates(&arr));
}
