use std::ops::{Add, Mul, Sub};
use std::thread;
use std::time::Instant;
use bytesize::ByteSize;
use futures::executor::block_on;
use rand::Rng;
use futures::future::{join_all};
use jemalloc_ctl::{stats, epoch};

#[global_allocator]
static ALLOC: jemallocator::Jemalloc = jemallocator::Jemalloc;

/// Only stores more efficiently when at least 50% of all elements are zeros
pub struct SparseVec {
    values: Vec<f64>,
    indices: Vec<usize>,
}

impl SparseVec {

    pub fn dot(&self, other: &SparseVec) -> f64 {
        let mut sum = 0.0;

        for index in 0..other.indices.len() {
            // exponential search for an element in the second vector to have the same index
            sum += binary_search(self.indices[index], &other.indices, &other.values) * self.values[index];
        }

        sum
    }

    pub fn new(elements: usize, non_null: f64) -> Self {
        let non_zero_elements = (elements as f64 * non_null) as usize;

        let mut values = Vec::with_capacity(non_zero_elements);
        let mut indices = Vec::with_capacity(non_zero_elements);

        let mut rng = rand::thread_rng();

        for i in 0..non_zero_elements {
            values.push(0.5);

            let idx = i as f32 / non_zero_elements as f32 * (elements as f32 - 4.0) + rng.gen_range(0.0..3.0);
            indices.push(idx as usize);
        }

        Self {
            values,
            indices
        }
    }
}

#[inline]
fn binary_search(target: usize, indices: &[usize], values: &[f64]) -> f64 {
    let mut range = 0..indices.len();
    loop {
        let mut median = (range.end - range.start) >> 1;
        if median == 0 {
            break;
        }
        median += range.start;

        if indices[median] == target {
            return values[median];
        }

        if indices[median] > target {
            range.end = median;
        } else {
            range.start = median;
        }
    }

    0.0
}

macro_rules! time {
    ($name:literal, $block:expr) => {{
        let start = Instant::now();
        $block;
        println!("{} took {}s", $name, start.elapsed().as_secs_f64());
    }}
}

fn main() {

    /// Theoretical size of the vector in elements
    /// This would mean the we would require 10 GBs of memory to store a single vector
    const VECTOR_SIZE: usize = 10_000_000_000;
    /// Ratio between null and non null elements per vector
    /// this means that only `NULL_NON_NULL_RATIO * 100%` elements of every vector will
    /// be non-null and thus actually stored
    const NULL_NON_NULL_RATIO: f64 = 0.02;

    let non_zero_elements = (VECTOR_SIZE as f64 * NULL_NON_NULL_RATIO) as usize;
    let heap_element_size = std::mem::size_of::<f64>() + std::mem::size_of::<usize>();

    println!("Estimated size on heap: {}", ByteSize::b((non_zero_elements * heap_element_size) as u64));
    println!("Size on stack: {} B", std::mem::size_of::<SparseVec>());

    let mut vec: SparseVec;

    time!("Sparse vector creation", {
        // generate a vector
        vec = SparseVec::new(VECTOR_SIZE, NULL_NON_NULL_RATIO);
    });

    // many statistics are cached and only updated when the epoch is advanced.
    epoch::advance().unwrap();
    println!("Heap allocated bytes (total): {}", ByteSize::b(stats::allocated::read().unwrap() as u64));

    time!("Sparse vector dot product", {
        vec.dot(&vec);
    });
}